@@ -54,6 +54,7 @@
 #include "qmi-voice.h"
 #include "qmi-loc.h"
 #include "qmi-qos.h"
+#include "qmi-gas.h"
 
 #define PACKED __attribute__((packed))
 
@@ -449,6 +450,53 @@ qmi_message_new (QmiService service,
     return (QmiMessage *)self;
 }
 
+QmiMessage *
+qmi_message_new_from_data (QmiService   service,
+                           guint8       client_id,
+                           GByteArray  *qmi_data,
+                           GError     **error)
+{
+    GByteArray *self;
+    struct full_message *buffer;
+    gsize buffer_len;
+    gsize message_len;
+
+    /* Create array with enough size for the QMUX marker and QMUX header, and
+     * with enough room to copy the rest of the message into */
+    if (service == QMI_SERVICE_CTL) {
+        message_len = sizeof (struct control_header) +
+            ((struct control_message *)(qmi_data->data))->header.tlv_length;
+    } else {
+        message_len = sizeof (struct service_header) +
+            ((struct service_message *)(qmi_data->data))->header.tlv_length;
+    }
+    buffer_len = (1 + sizeof (struct qmux) + message_len);
+    /* Create the GByteArray with buffer_len bytes preallocated */
+    self = g_byte_array_sized_new (buffer_len);
+    g_byte_array_set_size (self, buffer_len);
+
+    /* Set up fake QMUX header */
+    buffer = (struct full_message *)(self->data);
+    buffer->marker = QMI_MESSAGE_QMUX_MARKER;
+    buffer->qmux.length = GUINT16_TO_LE (buffer_len - 1);
+    buffer->qmux.flags = 0;
+    buffer->qmux.service = service;
+    buffer->qmux.client = client_id;
+
+    /* Move bytes from the qmi_data array to the newly created message */
+    memcpy (&buffer->qmi, qmi_data->data, message_len);
+    g_byte_array_remove_range (qmi_data, 0, message_len);
+
+    /* Check input message validity as soon as we create the QmiMessage */
+    if (!message_check (self, error)) {
+        /* Yes, we lose the whole message here */
+        qmi_message_unref (self);
+        return NULL;
+    }
+
+    return (QmiMessage *)self;
+}
+
 QmiMessage *
 qmi_message_response_new (QmiMessage       *request,
                           QmiProtocolError  error)
@@ -510,6 +558,22 @@ qmi_message_get_raw (QmiMessage *self,
     return self->data;
 }
 
+const guint8 *
+qmi_message_get_data (QmiMessage *self,
+                      gsize *length,
+                      GError **error)
+{
+    g_return_val_if_fail (self != NULL, NULL);
+    g_return_val_if_fail (length != NULL, NULL);
+
+    if (message_is_control (self))
+        *length = sizeof (struct control_header);
+    else
+        *length = sizeof (struct service_header);
+    *length += get_all_tlvs_length (self);
+    return (guint8 *)(&((struct full_message *)(self->data))->qmi);
+}
+
 /*****************************************************************************/
 /* TLV builder & writer */
 
@@ -1492,7 +1556,6 @@ qmi_message_get_printable_full (QmiMessage        *self,
     gchar *contents;
 
     g_return_val_if_fail (self != NULL, NULL);
-    g_return_val_if_fail (line_prefix != NULL, NULL);
 
     if (!line_prefix)
         line_prefix = "";
@@ -1558,6 +1621,9 @@ qmi_message_get_printable_full (QmiMessage        *self,
     case QMI_SERVICE_OMA:
         contents = __qmi_message_oma_get_printable (self, context, line_prefix);
         break;
+    case QMI_SERVICE_GAS:
+        contents = __qmi_message_gas_get_printable (self, context, line_prefix);
+        break;
     case QMI_SERVICE_WDA:
         contents = __qmi_message_wda_get_printable (self, context, line_prefix);
         break;
@@ -1619,14 +1685,40 @@ qmi_message_get_version_introduced_full (QmiMessage        *self,
     case QMI_SERVICE_OMA:
         return __qmi_message_oma_get_version_introduced (self, context, major, minor);
 
+    case QMI_SERVICE_GAS:
+        return __qmi_message_oma_get_version_introduced (self, context, major, minor);
+
     case QMI_SERVICE_WDA:
         return __qmi_message_wda_get_version_introduced (self, context, major, minor);
 
     case QMI_SERVICE_LOC:
         return __qmi_message_loc_get_version_introduced (self, context, major, minor);
 
+    case QMI_SERVICE_VOICE:
+        return __qmi_message_voice_get_version_introduced (self, context, major, minor);
+
+    case QMI_SERVICE_PDC:
+        return __qmi_message_pdc_get_version_introduced (self, context, major, minor);
+
+    case QMI_SERVICE_QOS:
+        return __qmi_message_qos_get_version_introduced (self, context, major, minor);
+
     default:
         /* For the still unsupported services, cannot do anything */
         return FALSE;
     }
 }
+
+gboolean
+__qmi_message_is_abortable (QmiMessage        *self,
+                            QmiMessageContext *context)
+{
+    switch (qmi_message_get_service (self)) {
+    case QMI_SERVICE_WDS:
+        return __qmi_message_wds_is_abortable (self, context);
+    case QMI_SERVICE_NAS:
+        return __qmi_message_nas_is_abortable (self, context);
+    default:
+        return FALSE;
+    }
+}
