@@ -30,6 +30,7 @@
 #include <stdint.h>
 #include <stdio.h>
 #include <pwd.h>
+#include <errno.h>
 
 #include "qmi-utils.h"
 #include "qmi-error-types.h"
@@ -121,14 +122,17 @@ __qmi_user_allowed (uid_t uid,
 /*****************************************************************************/
 
 gchar *
-__qmi_utils_get_driver (const gchar *cdc_wdm_path)
+__qmi_utils_get_driver (const gchar *cdc_wdm_path,
+                        GError **error)
 {
     static const gchar *subsystems[] = { "usbmisc", "usb" };
     guint i;
     gchar *device_basename;
     gchar *driver = NULL;
 
-    device_basename = g_path_get_basename (cdc_wdm_path);
+    device_basename = __qmi_utils_get_devname (cdc_wdm_path, error);
+    if (!device_basename)
+        return NULL;
 
     for (i = 0; !driver && i < G_N_ELEMENTS (subsystems); i++) {
         gchar *tmp;
@@ -155,6 +159,43 @@ __qmi_utils_get_driver (const gchar *cdc_wdm_path)
     return driver;
 }
 
+gchar *
+__qmi_utils_get_devpath (const gchar  *cdc_wdm_path,
+                         GError      **error)
+{
+    gchar *aux;
+
+    if (!g_file_test (cdc_wdm_path, G_FILE_TEST_IS_SYMLINK))
+        return g_strdup (cdc_wdm_path);
+
+    aux = realpath (cdc_wdm_path, NULL);
+    if (!aux) {
+        int saved_errno = errno;
+
+        g_set_error (error, QMI_CORE_ERROR, QMI_CORE_ERROR_FAILED,
+                     "Couldn't get realpath: %s", g_strerror (saved_errno));
+        return NULL;
+    }
+
+    return aux;
+}
+
+gchar *
+__qmi_utils_get_devname (const gchar  *cdc_wdm_path,
+                         GError      **error)
+{
+    gchar *aux;
+    gchar *devname = NULL;
+
+    aux = __qmi_utils_get_devpath (cdc_wdm_path, error);
+    if (aux) {
+        devname = g_path_get_basename (aux);
+        g_free (aux);
+    }
+
+    return devname;
+}
+
 /*****************************************************************************/
 
 static volatile gint __traces_enabled = FALSE;
